/*
	Generic linker script for Cortex-M STM32F4xx
	This file is included by the specific MCU linker script
 
	Use multi-region initializations for code, data, bss and noinit
*/
 
/*
	To set a function in RAM section:
	__attribute__ ((section(".text_RAM")))      void ramFunction (void) 

	To set an initialized data in a CCMRAM or BKPRAM:
	__attribute__ ((section(".data_CCMRAM")))	int32_t	nbData = 3 ;
	__attribute__ ((section(".data_BKPRAM")))	int32_t	nbData = 3 ;
	
	To set a zeroized data in CCMRAM or BKPRAM:
	__attribute__ ((section(".bss_CCMRAM")))	int32_t	nbData ;
	__attribute__ ((section(".bss_BKPRAM")))	int32_t	nbData ;

	To set an uninitialized data in RAM, CCMRAM or BKPRAM:
	__attribute__ ((section(".noinit")))	int32_t	nbData ;
	__attribute__ ((section(".noinit_CCMRAM")))	int32_t	nbData ;
	__attribute__ ((section(".noinit_BKPRAM")))	int32_t	nbData ;
*/

/*--------------------------------------------------------------------------
/* The main (privileged) stack definition, default heap definition */

/* Required amount of RAM for main stack */
Main_Stack_Size_ = 1024 ;

_estack_ = ORIGIN(RAM) + LENGTH(RAM);		/* End of main stack pointer just beyond the RAM */
_sstack_ = _estack_  - Main_Stack_Size_ ;	/* Start of main stack	*/

/* Default heap definitions.
 * The heap start immediately after the last statically allocated 
 * .bss/.noinit section, and extends up to the main stack start.
 */
Min_Heap_Size_ = 1024 ;					/* For RAM size check only */
PROVIDE ( _heap_begin_ = _end_noinit ) ;
PROVIDE ( _heap_end_   = _sstack_ ) ;

/* The entry point is informative since the Cortex-M reset vector points to it anyway */
ENTRY(start_)

/*--------------------------------------------------------------------------
 * Sections Definitions
 */

SECTIONS
{
    /*
     * The main stack address, IRQ vectors, the default IRQ handlers and the startup code
     * are stored in the .isr_vectors section, which goes to beginning of FLASH. 
     */
    .isr_vectors : ALIGN(4)
    {
        FILL(0xFF)
        
        KEEP(*(.isr_stackaddress))			/* Interrupt stack address */
        KEEP(*(.isr_vectors))				/* Interrupt vectors */
        *(.isr_handlers .isr_handlers.*)	/* Interrupt handlers */
        *(.startup .startup.*)				/* Startup code*/
		*(.after_vectors*)

    } >FLASH

    .inits : ALIGN (4)
    {
        /* 
         * Memory regions initialization tables.
         *
         * There are two kinds of arrays for each RAM region: one for data and one for bss.
         * The startup code uses these tables to perform the initialization.
         * 
         * The data table includes:
         * - from (LOADADDR()) (in FLASH)
         * - region_begin (ADDR())
         * - region_end (ADDR()+SIZEOF())
         *
         * The bss table includes:
         * - region_begin (ADDR())
         * - region_end (ADDR()+SIZEOF())
         *
         * WARNING: It is mandatory that the regions are word aligned, 
         * since the initialization code works only on words.
         */
         
         /* The startup code uses only the variables defined in this section */
         
        __data_regions_array_start = . ;
        
        /* Code in RAM */
        LONG(LOADADDR(.text_RAM)) ;
        LONG(ADDR(.text_RAM)) ;
        LONG(ADDR(.text_RAM)+SIZEOF(.text_RAM)) ;
        
        /* Initialized data in RAM */
        LONG(LOADADDR(.data)) ;
        LONG(ADDR(.data)) ;
        LONG(ADDR(.data)+SIZEOF(.data)) ;
        
        /* Initialized data in CCM RAM */
        LONG(LOADADDR(.data_CCMRAM)) ;
        LONG(ADDR(.data_CCMRAM)) ;
        LONG(ADDR(.data_CCMRAM)+SIZEOF(.data_CCMRAM)) ;
        
 		/* Backup RAM initialized data */
        LONG(LOADADDR(.data_BKPRAM)) ;
        LONG(ADDR(.data_BKPRAM)) ;
        LONG(ADDR(.data_BKPRAM)+SIZEOF(.data_BKPRAM)) ;

        __data_regions_array_end = . ;
        
        __bss_regions_array_start = . ;
        
        LONG(ADDR(.bss)) ;
        LONG(ADDR(.bss)+SIZEOF(.bss)) ;
        
        LONG(ADDR(.bss_CCMRAM)) ;
        LONG(ADDR(.bss_CCMRAM)+SIZEOF(.bss_CCMRAM)) ;
        
        LONG(ADDR(.bss_BKPRAM)) ;
        LONG(ADDR(.bss_BKPRAM)+SIZEOF(.bss_BKPRAM)) ;
        
        __bss_regions_array_end = . ;

		/*--------------------------------------------------------------------------
         * The C++ preinit code, i.e. an array of pointers to initialization 
         * functions to be performed before constructors.
         */
		. = ALIGN (4) ;
		PROVIDE_HIDDEN (__preinit_array_start = .) ;
		KEEP (*(.preinit_array .preinit_array.*))
		PROVIDE_HIDDEN (__preinit_array_end = .) ;

		/*--------------------------------------------------------------------------
         * The C++ init code, i.e. an array of pointers to static constructors.
         */
		. = ALIGN (4) ;
		PROVIDE_HIDDEN (__init_array_start = .) ;
		KEEP (*(SORT(.init_array.*)))
		KEEP (*(.init_array*))
		PROVIDE_HIDDEN (__init_array_end = .) ;

		/*--------------------------------------------------------------------------
         * The C++ fini code, i.e. an array of pointers to static destructors.
         * .fini_array ommited because when would we run destructors?
         */
		. = ALIGN (4) ;
		PROVIDE_HIDDEN (__fini_array_start = .) ;
		*(SORT(.fini_array.*))
		*(.fini_array)
		PROVIDE_HIDDEN (__fini_array_end = .) ;

    } >FLASH

    /*--------------------------------------------------------------------------
     * The program code is stored in the .text section, which goes to FLASH.
     */
	.text : ALIGN(4)
	{
		*(.text .text.*)			/* All the code */
 
		*(.rodata .rodata.* .constdata .constdata.*) 	/* Read-only data (constants) */		

		*(vtable)					/* C++ virtual tables */

		KEEP(*(.eh_frame*))

	} >FLASH

	/* ARM magic sections */
	.ARM.extab : ALIGN(4)
	{
		*(.ARM.extab* .gnu.linkonce.armextab.*)
	} > FLASH

	. = ALIGN(4);
	.ARM.exidx : ALIGN(4)
	{
		__exidx_start = .;   	
		*(.ARM.exidx* .gnu.linkonce.armexidx.*)
		__exidx_end = .;
	} > FLASH
   	
	. = ALIGN(4);
	_etext = .;
    
	/*--------------------------------------------------------------------------	
	 * The code section in RAM
	 */
	.text_RAM : ALIGN(4)
	{
       *(.text_RAM .text_RAM.*)
       . = ALIGN(4) ;
    } > RAM AT>FLASH

	/* --------------------------------------------------------------------------
     * The initialized data sections.
     */
	.data : ALIGN(4)
	{
		FILL(0xFF)

		*(.data)
		*(.data.*)

		. = ALIGN(4);
	} >RAM AT>FLASH

	/*  Some scripts uses these sections. Why ?  Warn if they are not empty */
	.begin_end :
	{
		*(.data_begin .data_begin.*)
		*(.data_end .data_end.*)
        *(.bss_begin .bss_begin.*)
        *(.bss_end .bss_end.*)
	} >RAM AT>FLASH
	ASSERT(SIZEOF(.begin_end) == 0, "begin_end != 0")

	.data_CCMRAM : ALIGN(4)
	{
		*(.data_CCMRAM .data_CCMRAM.*)
		. = ALIGN(4) ;
	} > CCMRAM AT>FLASH

	.data_BKPRAM : ALIGN(4)
	{
		*(.data_BKPRAM .data_BKPRAM.*)
		. = ALIGN(4) ;
	} >BKPRAM AT>FLASH

    /*--------------------------------------------------------------------------
     * The zeroized data sections. NOLOAD is used to avoid
     * the section type to be changed to PROGBITS
     */
     
	.bss_CCMRAM (NOLOAD) : ALIGN(4)
	{
		*(.bss_CCMRAM .bss_CCMRAM.*)
		. = ALIGN(4) ;
	} > CCMRAM

	.bss_BKPRAM (NOLOAD) : ALIGN(4)
	{
		*(.bss_BKPRAM .bss_BKPRAM.*)
		. = ALIGN(4) ;
	} > BKPRAM AT>FLASH

	/* The primary uninitialised data section. */
	.bss (NOLOAD) : ALIGN(4)
	{
		*(.bss)
		*(.bss.*)
		*(COMMON)
		. = ALIGN(4);
	} >RAM

    /*--------------------------------------------------------------------------
     * The uninitialized data sections. NOLOAD is used to avoid
     * the section type to be changed to PROGBITS
     */
    .noinit_CCMRAM (NOLOAD) : ALIGN(4)
    {
        *(.noinit_CCMRAM .noinit_CCMRAM.*)         
    } > CCMRAM
    
    .noinit_BKPRAM (NOLOAD) : ALIGN(4)
    {
        *(.noinit_BKPRAM .noinit_BKPRAM.*)         
         . = ALIGN(4) ;
    } > BKPRAM

	/* This section MUST be the last one in RAM */
	.noinit (NOLOAD) : ALIGN(4)
	{
		*(.noinit .noinit.*) 

		. = ALIGN(4) ;
		_end_noinit = . ;   /* Mandatory to be word aligned, _sbrk assumes this */
	} > RAM
    
	PROVIDE ( end  = _end_noinit );		/* For _sbrk, if used */
	PROVIDE ( _end = _end_noinit );
    
    /*--------------------------------------------------------------------------
     * Used for validation only, do not allocate anything in RAM from here
     *
     * This is to check that there is enough RAM left for the Main stack.
     * It generate an error if it's full.
     */
	._check_heap_stack : ALIGN(4)
	{
		. = . + Main_Stack_Size_ + Min_Heap_Size_ ;
	} >RAM

    /*--------------------------------------------------------------------------
    /*--------------------------------------------------------------------------
     * The FLASH Bank2
     * The C or assembly source must explicitly place the code 
     * or data there using the "section" attribute.
     */
    .text_b2 : ALIGN(4)
    {
        *(.text_b2*)					/* The code */
        *(.rodata_b2*)					/* Read-only data (constants) */
    } >FLASHB2
    
    /*--------------------------------------------------------------------------
    /*--------------------------------------------------------------------------
     * The EXTMEM.
     * The C or assembly source must explicitly place the code or data there
     * using the "section" attribute.
     */

	/* EXTMEM Bank0 -------------------------------------*/
	.data_eb0 : ALIGN(4)
	{
		*(.data_eb0*)				/* Initialized data */
	} >EXTMEMB0

	.data_eb0 : ALIGN(4)
	{
		*(.bss_eb0*)				/* Zeroized data */
	} >EXTMEMB0

	.data_eb0 (NOLOAD) : ALIGN(4)
	{
		*(.noinit_eb0*)				/* Non initialized data */
	} >EXTMEMB0

	/* EXTMEM Bank1 -------------------------------------*/
	.data_eb1 : ALIGN(4)
	{
		*(.data_eb1*)				/* Initialized data */
	} >EXTMEMB1

	.data_eb1 : ALIGN(4)
	{
		*(.bss_eb1*)				/* Zeroized data */
	} >EXTMEMB1

	.data_eb1 (NOLOAD) : ALIGN(4)
	{
		*(.noinit_eb1*)				/* Non initialized data */
	} >EXTMEMB1
    
	/* EXTMEM Bank2 -------------------------------------*/
	.data_eb2 : ALIGN(4)
	{
		*(.data_eb2*)				/* Initialized data */
	} >EXTMEMB2

	.data_eb2 : ALIGN(4)
	{
		*(.bss_eb2*)				/* Zeroized data */
	} >EXTMEMB2

	.data_eb2 (NOLOAD) : ALIGN(4)
	{
		*(.noinit_eb2*)				/* Non initialized data */
	} >EXTMEMB2
    
	/* EXTMEM Bank3 -------------------------------------*/
	.data_eb3 : ALIGN(4)
	{
		*(.data_eb3*)				/* Initialized data */
	} >EXTMEMB3

	.data_eb3 : ALIGN(4)
	{
		*(.bss_eb3*)				/* Zeroized data */
	} >EXTMEMB3

	.data_eb3 (NOLOAD) : ALIGN(4)
	{
		*(.noinit_eb3*)				/* Non initialized data */
	} >EXTMEMB3
    
	/*--------------------------------------------------------------------------*/
	/*--------------------------------------------------------------------------*/
    /* After that there are only debugging sections. */
    
    /* Remove the debugging information from the standard libraries */    
    DISCARD :
    {
     libc.a ( * )
     libm.a ( * )
     libgcc.a ( * )
    }

  
    /* Stabs debugging sections.  */
    .stab          0 : { *(.stab) }
    .stabstr       0 : { *(.stabstr) }
    .stab.excl     0 : { *(.stab.excl) }
    .stab.exclstr  0 : { *(.stab.exclstr) }
    .stab.index    0 : { *(.stab.index) }
    .stab.indexstr 0 : { *(.stab.indexstr) }
    .comment       0 : { *(.comment) }
 
    /*
     * DWARF debug sections.
     * Symbols in the DWARF debugging sections are relative to the beginning
     * of the section so we begin them at 0.  
     */

    /* DWARF 1 */
    .debug          0 : { *(.debug) }
    .line           0 : { *(.line) }

    /* GNU DWARF 1 extensions */
    .debug_srcinfo  0 : { *(.debug_srcinfo) }
    .debug_sfnames  0 : { *(.debug_sfnames) }

    /* DWARF 1.1 and DWARF 2 */
    .debug_aranges  0 : { *(.debug_aranges) }
    .debug_pubnames 0 : { *(.debug_pubnames) }

    /* DWARF 2 */
    .debug_info     0 : { *(.debug_info .gnu.linkonce.wi.*) }
    .debug_abbrev   0 : { *(.debug_abbrev) }
    .debug_line     0 : { *(.debug_line) }
    .debug_frame    0 : { *(.debug_frame) }
    .debug_str      0 : { *(.debug_str) }
    .debug_loc      0 : { *(.debug_loc) }
    .debug_macinfo  0 : { *(.debug_macinfo) }

    /* SGI/MIPS DWARF 2 extensions */
    .debug_weaknames 0 : { *(.debug_weaknames) }
    .debug_funcnames 0 : { *(.debug_funcnames) }
    .debug_typenames 0 : { *(.debug_typenames) }
    .debug_varnames  0 : { *(.debug_varnames) }    
}
